﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Collections;

namespace Editor.Collections
{
    public class GapBuffer<T>
    {
        private T[] singleItem;
        protected T[] buffer;
        protected int gapBeginOffset;
        protected int gapEndOffset;
        protected int gapLength;
        protected int minGapLength;
        protected int maxGapLength;

        public int Length
        {
            get { return buffer.Length - gapLength; }
        }

        public T this[int index]
        {
            get
            {
#if DEBUG
                if ( index < 0 || index >= Length) 
				    throw new ArgumentOutOfRangeException("index");
#endif			
			    return index < gapBeginOffset ? buffer[index] : buffer[index + gapLength];
            }
            set 
            {
#if DEBUG
                if( index < 0 || index >= Length ) 
				    throw new ArgumentOutOfRangeException("index");
#endif
                if( index < gapBeginOffset ) buffer[index] = value;
                else buffer[index + gapLength] = value;
            }
        }

        public GapBuffer()
        {
            singleItem = new T[1];
            minGapLength = 8;
            maxGapLength = 16;

            SetBuffer( new T[0] );
        }

        public GapBuffer( T[] array )
        {
            singleItem = new T[1];
            minGapLength = 8;
            maxGapLength = 16;

            SetBuffer( array );
        }

        public GapBuffer( T[] array, int minGapLength )
        {
            singleItem = new T[1];
            this.minGapLength = minGapLength;
            maxGapLength = minGapLength * 2;

            SetBuffer( array );
        }

        public T[] GetRange( int index, int length )
		{
#if DEBUG
			if( index < 0 || index > Length ) 
				throw new ArgumentOutOfRangeException("index");
			if( index < 0 || index + length > Length ) 
				throw new ArgumentOutOfRangeException("length");
#endif			
			int end = index + length;
            T[] array;

			if( end < gapBeginOffset ) 
            {
                array = new T[length]; 
                Array.Copy( buffer, index, array ,0, length );
                return array;
			}
			
			if ( index > gapBeginOffset) {
                array = new T[length]; 
                Array.Copy( buffer, index + gapLength, array ,0, length );
                return array;
			}

			array = new T[gapBeginOffset - index + end - gapBeginOffset];
            Array.Copy( buffer, index, array, 0, gapBeginOffset - index );
            Array.Copy( buffer, gapEndOffset, array, gapBeginOffset - index, end - gapBeginOffset );

            return array;
		}

        public void Insert( int index, T item )
		{
            singleItem[0] = item;

			Replace( index, 0, singleItem );
		}

        public void Insert( int index, T[] items )
		{
			Replace( index, 0, items );
		}
		
		public void Remove( int index, int length )
		{
			Replace( index, length, new T[0] );
		}

        public void Replace( int index, int length, T[] items )
        {
            if( items == null ) items = new T[0];
#if DEBUG
            if( index < 0 || index > Length )
                throw new ArgumentOutOfRangeException("index");
            if( length < 0 || index + length > Length )
                throw new ArgumentOutOfRangeException("length");
#endif
            PlaceGap( index, items.Length - length );
            gapEndOffset += length;
            Array.Copy( items, 0, buffer, gapBeginOffset, items.Length );
            gapBeginOffset += items.Length;
            gapLength = gapEndOffset - gapBeginOffset;

            if ( gapLength > maxGapLength ) MakeNewBuffer( gapBeginOffset, minGapLength );
        }

        public void SetBuffer( T[] array )
		{
			if ( array == null ) buffer = new T[0];
            else
            { 
                buffer = new T[array.Length];
                Array.Copy( array, buffer, array.Length );
            }

			gapBeginOffset = gapEndOffset = gapLength = 0;
		}

        private void PlaceGap( int newGapOffset, int minRequiredGapLength )
        {
            if( gapLength < minRequiredGapLength )
            {
                MakeNewBuffer( newGapOffset, minRequiredGapLength );
            }
            else
            {
                while( newGapOffset < gapBeginOffset )
                {
                    buffer[--gapEndOffset] = buffer[--gapBeginOffset];
                }
                while( newGapOffset > gapBeginOffset )
                {
                    buffer[gapBeginOffset++] = buffer[gapEndOffset++];
                }
            }
        }

        private void MakeNewBuffer( int newGapOffset, int newGapLength )
		{
			if ( newGapLength < minGapLength ) newGapLength = minGapLength;
			
			T[] newBuffer = new T[Length + newGapLength];
			
            if( newGapOffset < gapBeginOffset ) 
            {
				Array.Copy( buffer, 0, newBuffer, 0, newGapOffset );
				Array.Copy( buffer, newGapOffset, newBuffer, newGapOffset + newGapLength, gapBeginOffset - newGapOffset );
				Array.Copy( buffer, gapEndOffset, newBuffer, newBuffer.Length - ( buffer.Length - gapEndOffset ), buffer.Length - gapEndOffset );
			} 
            else 
            {
				Array.Copy( buffer, 0, newBuffer, 0, gapBeginOffset );
				Array.Copy( buffer, gapEndOffset, newBuffer, gapBeginOffset, newGapOffset - gapBeginOffset );
				int lastPartLength = newBuffer.Length - ( newGapOffset + newGapLength );
				Array.Copy( buffer, buffer.Length - lastPartLength, newBuffer, newGapOffset + newGapLength, lastPartLength );
			}
			
			gapBeginOffset = newGapOffset;
			gapEndOffset = newGapOffset + newGapLength;
			gapLength = newGapLength;
			buffer = newBuffer;
		}
    }
}
