﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.ComponentModel;
using Editor.Text;

namespace Editor.Project
{
    public sealed class UseCaseProject : TreeNode, IProjectItem
    {
        private EventArgs changeArgs;
        private string description;
        private BindingList<DescriptionItem> descriptionItems;
        private BindingList<DictionaryWord> dictionaryWords;
        private bool[] modelProperties;
        private bool objectChanging;
        private bool saved;
        private BindingList<SpecialStep> specialSteps;
        private IStyleTable styleTable;

        public IProjectItem this[int index]
        {
            get { return Nodes[index] as IProjectItem; }
        }

        public bool Composite
        {
            get { return true; }
        }

        public bool Changing
        {
            get { return objectChanging; }
        }

        public string Description
        {
            get { return description; }
            set { description = value; }
        }

        public BindingList<DescriptionItem> DescriptionItems
        {
            get { return descriptionItems; }
        }

        public BindingList<DictionaryWord> DictionaryWords
        {
            get { return dictionaryWords; }
        }

        public int ItemCount
        {
            get { return Nodes.Count; }
        }

        public string ItemName
        {
            get { return Text; }
            set 
            { 
                Text = value;

                ObjectChangedEvent( this, EventArgs.Empty );
            }
        }

        public bool[] ModelProperties
        {
            get { return modelProperties; }
            set 
            { 
                modelProperties = value; 

                ObjectChangedEvent( this, EventArgs.Empty );
            }
        }

        public IProjectItem ParentItem
        {
            get { return null; }
        }

        public bool Saved
        {
            get { return saved; }
            set
            {
                saved &= value;
            }
        }

        public BindingList<SpecialStep> SpecialSteps
        {
            get { return specialSteps; }
        }

        public IStyleTable StyleTable
        {
            get { return styleTable; }
        }

        public event EventHandler ObjectChanged;
        public event EventHandler ObjectChanging;

        public UseCaseProject( string name )
        {
            changeArgs = null;
            description = string.Empty;
            descriptionItems = new BindingList<DescriptionItem>();
            dictionaryWords = new BindingList<DictionaryWord>();
            ImageIndex = 0;
            modelProperties = new bool[20];
            objectChanging = false;
            saved = false;
            SelectedImageIndex = 0;
            specialSteps = new BindingList<SpecialStep>();
            styleTable = new StyleTable();
            Text = name;

            dictionaryWords.ListChanged += OnDictionaryChanged;
        }

        public void Add( IProjectItem item )
        {
            Insert( Nodes.Count, item );
        }

        public void BeginChange()
        {
            if( !objectChanging )
            {
                objectChanging = true;
                changeArgs = null;
            
                if( ObjectChanging != null ) ObjectChanging( this, EventArgs.Empty );
            }
        }

        public void Clear()
        {
            Nodes.Clear();

            ObjectChangedEvent( this, EventArgs.Empty );
        }

        public IProjectItem CreateSubItem( Arguments argumets )
        {
            if( argumets is UseCaseModelArguments )
            {
                UseCaseModel model = new UseCaseModel( (argumets as UseCaseModelArguments).Name, this );
                
                Add( model );

                return model;
            }

            return null;
        }

        public void EndChange( bool discardEvent )
        {
            if( objectChanging )
            {
                objectChanging = false;

                if( !discardEvent && changeArgs != null ) ObjectChangedEvent( this, changeArgs ); 
            }
        }

        public Delegate[] GetReferences()
        {
            return ObjectChanged.GetInvocationList();
        }

        public int IndexOf( IProjectItem item )
        {
            TreeNode itemNode = item as TreeNode;

            if( itemNode == null ) return -1;

            return Nodes.IndexOf( itemNode );
        }

        public int IndexOf( IProjectItem item, int startIndex )
        {
            for( int i = startIndex; i < Nodes.Count; i++ )
            {
                if( object.ReferenceEquals( item, Nodes[i] ) ) return i; 
            }

            return -1;
        }

        public void Insert( int index, IProjectItem item )
        {
            TreeNode itemNode = item as TreeNode;

            if( itemNode == null ) return;

            if( item is UseCaseModel )
            {
                item.ObjectChanged += OnProjectItemChanged;

                for( int i = 0; i < Nodes.Count; i++ )
                {
                    if( string.Compare( item.ItemName, Nodes[i].Text ) < 0 )
                    {
                        Nodes.Insert( i, itemNode );
                        return;
                    }
                }

                Nodes.Insert( Nodes.Count, itemNode );

                ObjectChangedEvent( this, EventArgs.Empty );
            }
        }

        public void InsertRange( int index, IProjectItem[] items )
        {
            throw new NotImplementedException();
        }

        public bool Load( Arguments arguments )
        {
            throw new NotImplementedException();
        }

        public bool Remove( IProjectItem item )
        {
            TreeNode itemNode = item as TreeNode;

            if( itemNode == null ) return false;

            item.ObjectChanged -= OnProjectItemChanged;

            Nodes.Remove( itemNode );

            ObjectChangedEvent( this, EventArgs.Empty );

            return true;
        }

        public void RemoveAt( int index )
        {
            IProjectItem item = Nodes[index] as IProjectItem;

            if( item != null ) item.ObjectChanged -= OnProjectItemChanged;

            Nodes.RemoveAt( index );

            ObjectChangedEvent( this, EventArgs.Empty );
        }

        public void RemoveRange( int index, int length )
        {
            throw new NotImplementedException();
        }
        
        public bool Save( Arguments arguments )
        {
            throw new NotImplementedException();
        }

        private void ObjectChangedEvent( object sender, EventArgs eventArgs )
        {
            if( objectChanging ) 
            {
                changeArgs = eventArgs;
                return;
            }

            if( ObjectChanged != null ) ObjectChanged( sender, eventArgs );
        }

        private void OnDictionaryChanged( object sender, ListChangedEventArgs eventArgs )
        {
            saved = false;

            ObjectChangedEvent( this, EventArgs.Empty );
        }

        private void OnProjectItemChanged( object sender, EventArgs eventArgs )
        {
            saved = false;

            ObjectChangedEvent( this, EventArgs.Empty );
        }
    }
}
