﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using Editor.View;
using Editor.Drawing;

namespace Editor.Text
{
    public sealed class Style : IStyle
    {
        private int ascent;
        private Color backColor;
        private bool bold;
        private EventArgs changeArgs;
        private int descent;
        private Font font;
        private FontFamily fontFamily;
        private int height;
        private bool italic;
        private string name;
        private bool objectChanging;
        private float size;
        private bool strikeout;
        private Color textColor;
        private bool underline;

        public Color BackColor
        {
            get { return backColor; }
            set 
            { 
                backColor = value;

                ObjectChangedEvent( this, EventArgs.Empty );
            }
        }

        public bool Bold 
        {
            get { return bold; }
            set 
            {
                bold = value; 

                ObjectChangedEvent( this, EventArgs.Empty );
            }
        }

        public bool Changing
        {
            get { return objectChanging; }
        }

        public Font Font
        {
            get 
            { 
                if( font == null ) 
                {
                    font = CreateFont();
                }

                return font;
            }
        }

        public FontFamily FontFamily
        {
            get { return fontFamily; }
            set 
            {
                fontFamily = value;
                
                ObjectChangedEvent( this, EventArgs.Empty );
            }
        }

        public bool Italic
        {
            get { return italic; }
            set 
            {
                italic = value;

                ObjectChangedEvent( this, EventArgs.Empty );
            }
        }

        public string Name
        {
            get { return name; }
            set { name = value; }
        }

        public float Size
        {
            get { return size; }
            set 
            {
                size = value;

                ObjectChangedEvent( this, EventArgs.Empty );
            }
        }

        public bool Strikeout
        {
            get { return strikeout; }
            set 
            {
                strikeout = value;

                ObjectChangedEvent( this, EventArgs.Empty );
            }
        }

        public Color TextColor
        {
            get { return textColor; }
            set 
            { 
                textColor = value;

                ObjectChangedEvent( this, EventArgs.Empty );
            }
        }

        public bool Underline
        {
            get { return underline; }
            set 
            {
                underline = value;
                
                ObjectChangedEvent( this, EventArgs.Empty );
            }
        }

        public event EventHandler ObjectChanged;
        public event EventHandler ObjectChanging;

        public Style()
        {
            ascent = -1;
            backColor = Color.White;
            bold = false;
            changeArgs = null;
            descent = -1;
            font = null;
            fontFamily = FontFamily.GenericSerif;
            height = -1;
            italic = false;
            name = "Generic Style";
            objectChanging = false;
            size = 16;
            strikeout = false;
            textColor = Color.Black;
            underline = false; 
        }

        public Style( string name )
        {
            ascent = -1;
            backColor = Color.White;
            bold = false;
            changeArgs = null;
            descent = -1;
            font = null;
            fontFamily = FontFamily.GenericSerif;
            height = -1;
            italic = false;
            this.name = name;
            objectChanging = false;
            size = 16;
            strikeout = false;
            textColor = Color.Black;
            underline = false;
        }

        public Style( string name, string fontFamily, float fontSize )
        {
            ascent = -1;
            backColor = Color.White;
            bold = false;
            changeArgs = null;
            descent = -1;
            font = null;
            this.fontFamily = new FontFamily( fontFamily );
            height = -1;
            italic = false;
            this.name = name;
            objectChanging = false;
            this.size = fontSize;
            strikeout = false;
            textColor = Color.Black;
            underline = false;
        }

        public IStyle Clone()
        {
            Style style = new Style( name, fontFamily.Name, size );

            style.BackColor = backColor;
            style.Bold = bold;
            style.Italic = italic;
            style.Strikeout = strikeout;
            style.TextColor = textColor;
            style.Underline = underline;

            return style;
        }

        public bool Equals( IStyle other )
        {
            if( object.ReferenceEquals( this, other ) ) return true;

            if( this.backColor != other.BackColor ) return false;
            if( this.bold != other.Bold ) return false;
            if( string.Compare( this.fontFamily.Name, other.FontFamily.Name ) != 0 ) return false;
            if( this.italic != other.Italic) return false;
            if( this.size != other.Size ) return false;
            if( this.strikeout != other.Strikeout ) return false;
            if( this.textColor != other.TextColor ) return false;
            if( this.underline != other.Underline ) return false;
            
            return true;
        }

        public void BeginChange()
        {
            if( !objectChanging )
            {
                objectChanging = true;
                changeArgs = null;

                if( ObjectChanging != null ) ObjectChanging( this, EventArgs.Empty );
            }
        }

        private Font CreateFont()
        {
            FontStyle fontStyle = FontStyle.Regular;

            if( bold ) fontStyle = fontStyle | FontStyle.Bold;
            if( italic ) fontStyle = fontStyle | FontStyle.Italic;
            if( strikeout ) fontStyle = fontStyle | FontStyle.Strikeout;
            if( underline ) fontStyle = fontStyle | FontStyle.Underline;

            return new Font( fontFamily.Name, size, fontStyle, GraphicsUnit.Pixel );
        }

        public void EndChange( bool discardEvent )
        {
            if( objectChanging )
            {
                objectChanging = false;

                if( !discardEvent && changeArgs != null ) ObjectChangedEvent( this, changeArgs ); 
            }
        }

        public Delegate[] GetReferences()
        {
            return ObjectChanged.GetInvocationList();
        }

        public int GetAscent( IView view )
        {
            if( ascent == -1 ) font = CreateFont();
                CalculateMetrics( view );

            return ascent;
        }

        public int GetDescent( IView view )
        {
            if( descent == -1 ) font = CreateFont();
                CalculateMetrics( view );

            return descent;
        }

        public int GetHeight( IView view )
        {
            if( height == -1 ) font = CreateFont();
                CalculateMetrics( view );

            return height;
        }

        private void CalculateMetrics( IView view )
        {
            TextMetrics textMetrics = Gdi.GetTextMetrics( view.Graphics, Font );
            
            ascent = textMetrics.tmAscent;
            descent = textMetrics.tmDescent;
            height = textMetrics.tmHeight;
        }

        private void ObjectChangedEvent( object sender, EventArgs eventArgs )
        {
            ResetFont();

            if( objectChanging ) 
            {
                changeArgs = eventArgs;
                return;
            }

            if( ObjectChanged != null ) ObjectChanged( sender, eventArgs );
        }

        private void ResetFont()
        {
            ascent = -1;
            descent = -1;
            height = -1;
            font = null;
        }
    }
}
