﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.Drawing;

namespace Editor.Text
{
    public sealed class StyleTable : IStyleTable
    {
        private EventArgs changeArgs;
        private readonly short defaultStyles;
        private bool objectChanging;
        private Dictionary<short, IStyle> styleTable;
        
        public IStyle this[short index]
        {
            get 
            {
                if( !styleTable.ContainsKey( index ) )
                {
                    if( IsDefault( index ) ) styleTable[index] = new Style();
                }
                
                return styleTable[index];      
            }
            set 
            {
                styleTable[index].ObjectChanged -= OnStyleChanged;

                styleTable[index] = value;

                styleTable[index].ObjectChanged += OnStyleChanged;
            }
        }

        public bool Changing
        {
            get { return objectChanging; }
        }

        public short DefaultStyles 
        {
            get { return defaultStyles; }
        }

        public short MaximumIndex 
        {
            get { return 64; }
        }

        public event EventHandler ObjectChanged;
        public event EventHandler ObjectChanging;

        public StyleTable()
        {
            changeArgs = null;
            defaultStyles = Convert.ToInt16( Enum.GetValues( typeof(EDefaultStyles) ).Length );
            styleTable = new Dictionary<short, IStyle>();
            objectChanging = false;

            InitializeDefaults();
        }

        public void Add( short index, IStyle item )
        {
            if( index > MaximumIndex ) throw new ArgumentOutOfRangeException("index");

            item.ObjectChanged += OnStyleChanged;

            styleTable.Add( index, item );
        }

         public void BeginChange()
        {
            if( !objectChanging )
            {
                objectChanging = true;
                changeArgs = null;

                if( ObjectChanging != null ) ObjectChanging( this, EventArgs.Empty );
            }
        }

        public void Clear()
        {
            for( short i = defaultStyles; i < MaximumIndex; i++ )
            {
                if( styleTable.ContainsKey( i ) )
                {
                    styleTable[i].ObjectChanged -= OnStyleChanged;

                    styleTable.Remove( i );          
                }
            }
        }

        
        public bool Contains( short index )
        {
            if( index > MaximumIndex ) throw new ArgumentOutOfRangeException("index");
            if( IsDefault( index ) ) return true;

            return styleTable.ContainsKey( index );
        }

        public void EndChange( bool discardEvent )
        {
            if( objectChanging )
            {
                objectChanging = false;

                if( !discardEvent && changeArgs != null ) ObjectChangedEvent( this, changeArgs ); 
            }
        }

        public Delegate[] GetReferences()
        {
            return ObjectChanged.GetInvocationList();
        }

        public IEnumerator GetEnumerator()
        {
            return styleTable.Keys.GetEnumerator();
        }

        public bool IsDefault( short index )
        {
            if( index < defaultStyles ) return true;

            return false;
        }

        public bool Remove( short index )
        {
            if( index > MaximumIndex ) return false;

            styleTable[index].ObjectChanged -= OnStyleChanged;

            return styleTable.Remove( index );
        }

        private void InitializeDefaults()
        {
            Style style = new Style("Normal");

            styleTable.Add( (short)EDefaultStyles.Normal, style );
            
            style = new Style("Dictionary Word");
            style.Bold = true;
            style.TextColor = Color.Green;

            styleTable.Add( (short)EDefaultStyles.Definition, style );
                        
            style = new Style("Actor");
            style.Bold = true;
            style.TextColor = Color.Red;

            styleTable.Add( (short)EDefaultStyles.Actor, style );

            style = new Style("Document Section");
            style.Size = 18;
            style.Bold = true;
            style.Italic = true;

            styleTable.Add( (short)EDefaultStyles.Section, style );

            style = new Style("Flow Name");
            style.Bold = true;

            styleTable.Add( (short)EDefaultStyles.FlowName, style );

            style = new Style("Flow Numbering");
            style.Bold = true;

            styleTable.Add( (short)EDefaultStyles.FlowNumbering, style );

            style = new Style("Flow Link");
            style.Underline = true;
            style.TextColor = Color.Blue;

            styleTable.Add( (short)EDefaultStyles.FlowLink, style );

            style = new Style("Step Numbering");

            styleTable.Add( (short)EDefaultStyles.StepNumbering, style );

            style = new Style("Step Link");
            style.Underline = true;
            style.TextColor = Color.Blue;

            styleTable.Add( (short)EDefaultStyles.StepLink, style );
        }

        private void ObjectChangedEvent( object sender, EventArgs eventArgs )
        {
            if( objectChanging ) 
            {
                changeArgs = eventArgs;
                return;
            }

            if( ObjectChanged != null ) ObjectChanged( sender, eventArgs );
        }

        private void OnStyleChanged( object sender, EventArgs eventArgs )
        {
            ObjectChangedEvent( sender, EventArgs.Empty );
        }
    }
}
