﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using Editor.Document.UseCase;
using Editor.Document;
using System.Drawing.Text;
using Editor.Text;

namespace Editor.View
{
    public sealed class UseCaseView : IView
    {
        private Caret caret;
        private EventArgs changeArgs;
        private Graphics controlGraphics;
        private IDocument document;
        private UseCaseEditor editor;
        private Graphics drawGraphics;
        private int rowLevelIndent;
        private int margins;
        private Size maximumSize;
        private bool mouseDown;
        private bool objectChanging;
        private Size size;
        private IStyleTable styleTable;
        private const TextFormatFlags textFormatFlags = TextFormatFlags.NoPadding | 
            TextFormatFlags.NoPrefix | TextFormatFlags.PreserveGraphicsClipping | 
            TextFormatFlags.PreserveGraphicsTranslateTransform;
        private Dictionary<IDocumentRow, IViewRow> viewRowTable;

        public IViewRow this[int index]
        {
            get { return viewRowTable[document[index]]; }
        }

        public Caret Caret 
        {
            get { return caret; } 
        }

        public bool Changing
        {
            get { return objectChanging; }
        }

        public IDocument Document
        {
            get { return document; }
            set 
            {
                caret.Hide();

                document = value;

                PlaceCaret( int.MaxValue );
            }
        }

        public Graphics Graphics
        {
            get 
            {
                if( drawGraphics == null ) return controlGraphics;

                return drawGraphics;
            }
        }

        public int RowLevelIndent 
        {
            get { return rowLevelIndent; }
            set
            {
                rowLevelIndent = value;

                PrepareAll();

                ObjectChangedEvent( this, EventArgs.Empty );
            }
        }

        public int Margins
        {
            get { return margins; }
            set
            {
                margins = value;

                PrepareAll();

                ObjectChangedEvent( this, EventArgs.Empty );
            }
        }

        public int RowCount
        {
            get { return document.RowCount; }
        }

        public Size Size
        {
            get { return size; }
            set 
            { 
                size = value;

                PrepareAll();

                ObjectChangedEvent( this, EventArgs.Empty );
            }
        }

        public IStyleTable StyleTable 
        {
            get { return styleTable; }
        }

        public event EventHandler ObjectChanged;
        public event EventHandler ObjectChanging;

        public UseCaseView( UseCaseEditor editor, UseCaseDocument document )
        {
            caret = new Caret( this, 500, Color.Black );
            changeArgs = null;
            controlGraphics = editor.CreateGraphics();
            this.document = document;
            drawGraphics = null;
            this.editor = editor;
            this.rowLevelIndent = 20;
            this.margins = 5;
            maximumSize = new Size( int.MaxValue, int.MaxValue );
            mouseDown = false;
            objectChanging = false;
            styleTable = document.Project.StyleTable;
            size.Height = 0;
            size.Width = 450;
            viewRowTable = new Dictionary<IDocumentRow, IViewRow>();
  
            InitializeRows();

            editor.MouseDown += new MouseEventHandler( OnMouseDown );
            editor.MouseUp += new MouseEventHandler( OnMouseUp );
            editor.MouseMove += new MouseEventHandler( OnMouseMove );

            document.RowAdded += OnRowAdded;
            document.RowRemoved += OnRowRemoved;
            document.ObjectChanged += OnDocumentChanged;
            document.ObjectChanging += OnDocumentChanging;
            
            styleTable.ObjectChanged += OnStyleTableChanged;
        }

        public void BeginChange()
        {
            if( !objectChanging )
            {
                objectChanging = true;
                changeArgs = null;

                if( ObjectChanging != null ) ObjectChanging( this, EventArgs.Empty );
            }
        }

        public void Draw( Arguments arguments )
        {
            PaintArguments paintArguments = arguments as PaintArguments;
            drawGraphics = paintArguments.Graphics;
            
            for( int i = 0; i < document.RowCount; i++ )       
                viewRowTable[document[i]].Draw();

            drawGraphics = null;
        }

        public void DrawString( string text, Font font, Color textColor, Point point )
        {
            TextRenderer.DrawText( drawGraphics, text, font, point, textColor, textFormatFlags );
        }

        public void DrawString( string text, Font font, Color textColor, Color backColor, Point point )
        {
            TextRenderer.DrawText( drawGraphics, text, font, point, textColor, backColor, textFormatFlags );
        }
        
        public void EndChange( bool discardEvent )
        {
            if( objectChanging )
            {
                objectChanging = false;

                if( !discardEvent && changeArgs != null ) ObjectChangedEvent( this, changeArgs ); 
            }
        }

        public Delegate[] GetReferences()
        {
            return ObjectChanged.GetInvocationList();
        }

        public int GetSelectionRow( int y )
        {
            if( document.RowCount == 0 ) return -1;

            if( y < margins ) return 0;

            IViewRow viewRow = null;

            for( int i = 0; i < document.RowCount; i++ )
            {
                viewRow = viewRowTable[document[i]];

                if( viewRow.Offset + viewRow.Height > y ) return i;
            }

            return document.RowCount - 1;
        }

        public Size MeasureString( string text, Font font )
        {
            return TextRenderer.MeasureText( controlGraphics, text, font, maximumSize, textFormatFlags );
        }

        public void PlaceCaret( int y )
        {
            if( document.RowCount == 0 ) return;

            SSelection selection = document.Selection.GetCurrent();
            IViewRow endRow = viewRowTable[document[selection.EndRow]];

            endRow.PlaceCaret( selection.EndIndex, y, document.Selection.Reversed ? -1 : 0 );
        }

        public void Prepare()
        {
            int offset = margins;
            
            IViewRow viewRow = null;

            for( int i = 0; i < document.RowCount; i++ )
            {
                viewRow = viewRowTable[document[i]];

                viewRow.Offset = offset;
                offset += viewRow.Height;
            }

            size.Height = offset - margins;
        }

        public void Redraw()
        {
            editor.Invalidate();
        }

        private void InitializeRows()
        {
            IViewRow createdRow = null;

            for( int i = 0; i < document.RowCount; i++ )
            {
                createdRow = CreateRow( document[i] );

                viewRowTable.Add( createdRow.DocumentRow, createdRow );
            }
        }

        private void ObjectChangedEvent( object sender, EventArgs eventArgs )
        {
            if( objectChanging ) 
            {
                changeArgs = eventArgs;
                return;
            }

            if( ObjectChanged != null ) ObjectChanged( sender, eventArgs );      
        }

        private void OnDocumentChanged( object sender, EventArgs eventArgs )
        {
            //CorrectSelection();

            Prepare();

            PlaceCaret( int.MaxValue );

            ObjectChangedEvent( sender, eventArgs );
        }

        private void CorrectSelection()
        {
            if( document.RowCount > 0 )
            {
                if( document.Selection.StartRow >= document.RowCount || 
                    document.Selection.EndRow >= document.RowCount )
                {
                    IDocumentRow lastRow = document[document.RowCount - 1];

                    document.Selection.Change(document.RowCount - 1, lastRow.Text.CharacterLength,
                        document.RowCount - 1, lastRow.Text.CharacterLength);
                }
            }
        }

        private void OnDocumentChanging( object sender, EventArgs eventArgs )
        {
            caret.Hide();
        }

        private void OnMouseDown( object sender, MouseEventArgs mouseEventArgs )
        {
            mouseDown = true;

            Point location = mouseEventArgs.Location;
            location.X -= editor.AutoScrollPosition.X;
            location.Y -= editor.AutoScrollPosition.Y;

            caret.Hide();

            int selectionRow = GetSelectionRow( location.Y );

            if( selectionRow == -1 )
                document.Selection.Change( 0, 0, 0, 0 );
            else
            {
                int selectionIndex = this[selectionRow].GetSelectionIndex( location.X, 
                    location.Y );

                document.Selection.Change( selectionRow, selectionIndex, selectionRow,
                    selectionIndex );
            }

            Redraw();
        }

        private void OnMouseMove( object sender, MouseEventArgs mouseEventArgs )
        {
            if ( mouseDown == true ) 
            {
                Point location = mouseEventArgs.Location;
                location.X -= editor.AutoScrollPosition.X;
                location.Y -= editor.AutoScrollPosition.Y;
  
                int selectionRow = GetSelectionRow( location.Y );

                if( selectionRow == -1 )
                    document.Selection.Change( 0, 0, 0, 0 );
                else
                {
                    int selectionIndex = this[selectionRow].GetSelectionIndex(location.X, 
                        location.Y );

                    document.Selection.Change( document.Selection.StartRow, document.Selection.StartIndex, 
                        selectionRow, selectionIndex );
                }
            }

            Redraw();
        }

        private void OnMouseUp( object sender, MouseEventArgs mouseEventArgs )
        {
            mouseDown = false;

            Point location = mouseEventArgs.Location;
            location.X -= editor.AutoScrollPosition.X;
            location.Y -= editor.AutoScrollPosition.Y;

            int selectionRow = GetSelectionRow( location.Y );

            if( selectionRow == -1 )
                document.Selection.Change( -1, -1, -1, -1 );
            else
            {
                int selectionIndex = this[selectionRow].GetSelectionIndex( location.X, 
                    location.Y );

                document.Selection.Change( document.Selection.StartRow,
                    document.Selection.StartIndex, selectionRow, selectionIndex );
            }         

            PlaceCaret( location.Y );

            Redraw();
        }

        private void OnRowAdded( object sender, EventArgs eventArgs )
        {
            IViewRow createdRow = CreateRow( sender as IDocumentRow );

            viewRowTable.Add( createdRow.DocumentRow, createdRow );
        }

        private void OnRowRemoved( object sender, EventArgs eventArgs )
        {
            viewRowTable.Remove( sender as IDocumentRow );
        }

        private void OnStyleTableChanged( object sender, EventArgs eventArgs )
        {
            PrepareAll();

            ObjectChangedEvent( sender, eventArgs );
        }

        private void PrepareAll()
        {
            IViewRow viewRow = null;

            for( int i = 0; i < document.RowCount; i++ )
            {
                viewRow = viewRowTable[document[i]];

                viewRow.Prepare();
            } 
        }

        private IViewRow CreateRow( IDocumentRow documentRow )
        {
            return new ViewRow( this, documentRow );
        }
    }
}
